#ifndef MOVINGOBJECTBASE_H
#define MOVINGOBJECTBASE_H

#include "ObjectBase.h"


//----------------------------------------------------------------------------
/**
 * Base class for all objects that move,
 * such as ships, projectiles, particles, etc.
 */
class MovingObjectBase : public ObjectBase
{
  public:
    virtual ~MovingObjectBase();

    /**
     * Update the x and y values for m_position
     * out m_fineXPosition and m_fineYPosition.
     */
    inline void updatePosition()
    {
        ObjectBase::setPosition((Sint16)rint(m_fineXPosition),
                                (Sint16)rint(m_fineYPosition));
    }

    /**
     * Overload ObjectBase::setPosition() to set the fine-scale position
     * stored in m_fineXPosition and m_fineYPosition.
     *
     * @param x The new x position.
     * @param y The new y position.
     */
    inline void setPosition(Sint16 x, Sint16 y)
    {
        m_fineXPosition = 1.0 * x;
        m_fineYPosition = 1.0 * y;
        ObjectBase::setPosition(x, y);
    }


    /**
     * @return The current fine scale x position.
     */
    inline double getFineXPosition() const
    {
        return m_fineXPosition;
    }

    /**
     * @return The current fine scale y position.
     */
    inline double getFineYPosition() const
    {
        return m_fineYPosition;
    }


    inline void incFineXPosition(double increment)
    {
        m_fineXPosition += increment;
    }
    
    inline void incFineYPosition(double increment)
    {
        m_fineYPosition += increment;
    }

    inline void decFineXPosition(double decrement)
    {
        m_fineXPosition -= decrement;
    }
    
    inline void decFineYPosition(double decrement)
    {
        m_fineYPosition -= decrement;
    }


    /**
     * @return The current fine scale x velocity.
     */
    inline double getFineXVelocity() const
    {
        return m_fineXVelocity;
    }

    /**
     * @return The current fine scale y velocity.
     */
    inline double getFineYVelocity() const
    {
        return m_fineYVelocity;
    }

    inline void setFineXVelocity(double x)
    {
        m_fineXVelocity = x;
    }

    inline void setFineYVelocity(double y)
    {
        m_fineYVelocity = y;
    }

    inline void setFineVelocity(double x, double y)
    {
        setFineXVelocity(x);
        setFineYVelocity(y);
    }

    inline void incFineXVelocity(double increment)
    {
        m_fineXVelocity += increment;
    }
    
    inline void incFineYVelocity(double increment)
    {
        m_fineYVelocity += increment;
    }

    inline void decFineXVelocity(double decrement)
    {
        m_fineXVelocity -= decrement;
    }
    
    inline void decFineYVelocity(double decrement)
    {
        m_fineYVelocity -= decrement;
    }


    /**
     * This update() implementation will call do_update().
     * The object's position before and after the do_update() call
     * will be merged and stored in updateRect.
     *
     * @param updateRect The update rectangle.
     */
    void update(SDL_Rect &updateRect);

  protected:

    MovingObjectBase(const ObjectBase *creator);

  private:

    /// Called by update() to do the actual updating of the object.
    virtual void do_update() = 0;


    /// The current fine-scale x-velocity of this object.
    double m_fineXVelocity;

    /// The current fine-scale y-velocity of this object.
    double m_fineYVelocity;


    /// The object's fine-scale x position.
    double m_fineXPosition;

    /// The object's fine-scale y position.
    double m_fineYPosition;
};

#endif //MOVINGOBJECTBASE_H
