#ifndef TURRETS_H
#define TURRETS_H

#include <list>

#include "SinusTable.h"
#include "Tools.h"

#include "XMLVisitors.h"

#include "StaticDecorationBase.h"


//----------------------------------------------------------------------------
class TurretSurfaces : public OrientatingSurfacesBase
{
    SURFACES_SINGLETON_OBJECT(TurretSurfaces);
};


//----------------------------------------------------------------------------
class TurretBase;


//----------------------------------------------------------------------------
#undef DECLARE_UPDATE_STRATEGY
#define DECLARE_UPDATE_STRATEGY(c) \
    class c; \
    friend class c; \
    class c : public UpdateStrategy \
    { \
        STRATEGY_OBJECT(c); \
      public: \
        c() {} \
        ~c() {} \
        void update(TurretBase *t, Barrel *b); \
    }


//----------------------------------------------------------------------------
class Barrel
{
    //------------------------------------------------------------------------
    enum EnumUpdateStrategy
    {
        US_FIXED,
        US_RANDOM,
        US_SWEEP,
        US_SMART
    };

    //------------------------------------------------------------------------
    class UpdateStrategy
    {
      public:
        UpdateStrategy() {}
        virtual ~UpdateStrategy() {}

        static UpdateStrategy *get(EnumUpdateStrategy strategy);

        virtual void update(TurretBase *t, Barrel *b) = 0;
    };

    DECLARE_UPDATE_STRATEGY(Fixed);
    DECLARE_UPDATE_STRATEGY(Random);
    DECLARE_UPDATE_STRATEGY(Sweep);
    DECLARE_UPDATE_STRATEGY(Smart);

  public:

    //------------------------------------------------------------------------
    class InitializationData
    {
      public:
        InitializationData(const XMLNode *barrelNode, TurretBase *t)
            throw (XMLException);
        ~InitializationData();

        TurretBase *turret;

        EnumUpdateStrategy updateStrategy;

        int angle;
        unsigned step;

        unsigned speed;
        unsigned delay;
    };

    ~Barrel();

    inline int getAngle() const
    {
        return m_angle;
    }
    inline void setAngle(int angle)
    {
        m_angle = SinusTable::normalize(angle);
    }

    inline unsigned getAngleStep() const
    {
        return m_angleStep;
    }

    inline unsigned getProjectileSpeed() const
    {
        return m_projectileSpeed;
    }

    void update(TurretBase *turret);

    inline void setStrategy(UpdateStrategy *strategy)
    {
        m_strategy = strategy;
    }

    static Barrel *create(const XMLNode *barrelNode, TurretBase *turret)
        throw (Exception);

  protected:
    Barrel(const InitializationData &init);

    TurretBase *m_turret;

    int m_angle;
    unsigned m_angleStep;

    unsigned m_projectileSpeed;

    unsigned m_frameDelay;
    unsigned m_frameCounter;

  private:
    UpdateStrategy *m_strategy;
};

typedef std::list<Barrel*> BarrelList;
typedef BarrelList::iterator BarrelListIter;
typedef BarrelList::const_iterator BarrelListCIter;



//----------------------------------------------------------------------------
class TurretBase : public StaticDecorationBase,
                   public OrientatingDecorationBase
{
    //------------------------------------------------------------------------
    class BarrelVisitor;
    friend class BarrelVisitor;
    class BarrelVisitor : public XMLConstVisitor
    {
      public:
        //--------------------------------------------------------------------
        BarrelVisitor(TurretBase *turret) : m_turret(turret) {}
        ~BarrelVisitor() { m_turret = NULL; }

      private:
        //--------------------------------------------------------------------
        void do_visit(const XMLProperty *a);
        void do_visit(const XMLNode *n);

        //--------------------------------------------------------------------
        TurretBase *m_turret;
    };

  public:
    //------------------------------------------------------------------------
    class InitializationData
        : public StaticDecorationBase::InitializationData,
          public OrientatingDecorationBase::InitializationData
    {
      public:
        InitializationData(const XMLNode *turretNode)
            throw (XMLException);
        ~InitializationData();

        unsigned hitPoints;
    };


    virtual ~TurretBase();

    static TurretBase *create(const XMLNode *turretNode)
        throw (Exception);

    /// Returns, whether the turret can fire to the given angle.
    bool isAngleInRange(int angle) const;
    virtual int getCenterAngle() const = 0;

    //------------------------------------------------------------------------
    /**
     * Called for a projectile collision to determine,
     * if the turret's hitpoints are reached.
     */
    inline bool decAndTestHitPoints()
    {
        return --m_hitPoints == 0;
    }

    /**
     * Creates some explosion particles on the turret's position.
     */
    void createExplosionParticles();

  protected:
    TurretBase(const InitializationData &init);

    static TurretBase *create(const InitializationData &init);

  private:
    //------------------------------------------------------------------------
    void do_update();

    //------------------------------------------------------------------------
    DECLARE_VISITOR_API();

    BarrelList m_barrels;

    unsigned m_hitPoints;
};


//----------------------------------------------------------------------------
class TopTurret : public TurretBase
{
  public:
    TopTurret(const InitializationData &init);
    ~TopTurret();

    int getCenterAngle() const;
};

//----------------------------------------------------------------------------
class BottomTurret : public TurretBase
{
  public:
    BottomTurret(const InitializationData &init);
    ~BottomTurret();

    int getCenterAngle() const;
};

//----------------------------------------------------------------------------
class LeftTurret : public TurretBase
{
  public:
    LeftTurret(const InitializationData &init);
    ~LeftTurret();

    int getCenterAngle() const;
};

//----------------------------------------------------------------------------
class RightTurret : public TurretBase
{
  public:
    RightTurret(const InitializationData &init);
    ~RightTurret();

    int getCenterAngle() const;
};

#endif //TURRETS_H
